
from pycloak import auth, admin

def test_get_valid_config(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='first broker login').executions()
    auth_config = next(filter(lambda execution: execution.get('authenticationConfig') is not None, executions), None)
    assert auth_config is not None, "couldn't find a configurable execution"
    print(auth_config['authenticationConfig'])
    assert master_realm.auth_config(auth_config['authenticationConfig']) is not None

def test_create_config(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = {'alias': 'test create config flow', 'providerId': 'basic-flow', 'description': 'This flow is used for testing config creation', 'topLevel': 'true', 'builtIn': 'false'}
    created_auth_flow = master_realm.create_auth_flow(auth_flow)
    idp_redirector = {'provider': 'identity-provider-redirector'}
    created_execution = created_auth_flow.create_execution(idp_redirector)
    new_config = created_execution.create_config({'alias': 'test create config', 'config': {'defaultProvider': 'https://www.github.com'}})
    assert new_config is not None




def merge(preferred, secondary):
    if isinstance(preferred, dict):
        if not isinstance(secondary, dict):
            return preferred
        return mergeDicts(preferred, secondary)

    if isinstance(preferred, list):
        if not isinstance(secondary, list):
            return preferred

        return mergeLists(preferred, secondary)
    else:
        return preferred

def mergeDicts(preferred, secondary):
    return_dict = {}

    for key, preferred_value in preferred.items():
        if secondary.get(key) is None:
            return_dict[key] = preferred_value
        else:
            return_dict[key] = merge(preferred_value, secondary[key])

    # Perhaps not the most efficient, but you know what they say about premature optimization....
    for key, secondary_value in secondary.items():
        if return_dict.get(key) is None:
            return_dict[key] = secondary_value

    return return_dict

def mergeLists(preferred, secondary):
    return preferred




import json
import requests
from pycloak import auth, admin, flow

class Execution:

    def __init__(self, flow, json_rep=None, dict_rep=None):
        # TODO pretty sure I don't have to do this.  Explore what python typing automagically does for me.
        if not json_rep and dict_rep:
            json_rep = json.loads(json.dumps(dict_rep))

        self.flow = flow
        self.id = json_rep.get('id')
        self.json = json_rep

    def create_config(self, config):
        url = "{0}/auth/admin/realms/{1}/authentication/executions/{2}/config".format(self.flow.realm.auth_session.host, self.flow.realm.id, self.id)
        response = requests.post(url, json=config, headers=self.flow.realm.auth_session.bearer_header)

        if response.status_code != 201:
            raise KeycloakExecutionException("Error attempting to create new execution: {0}/{1}".format(response.status_code, response.text))

        return requests.get(response.headers['Location']).text

# TODO too many exceptions, clean up handling
class KeycloakExecutionException(Exception):
    pass




from pycloak import auth, admin, merge
import pytest
import json


def test_get_auth_flows(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    auth_flows = admin.Admin(session).realm('master').auth_flows()
    assert auth_flows is not None

def test_create_auth_flow(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    auth_flow = {'alias': 'test flow', 'providerId': 'basic-flow', 'description': 'This flow is used for test purposes', 'topLevel': 'true', 'builtIn': 'false'}
    created_auth_flow = admin.Admin(session).realm('master').create_auth_flow(json.loads(json.dumps(auth_flow)))
    assert created_auth_flow is not None

def test_get_auth_flow(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = master_realm.auth_flow(alias='test flow')
    assert auth_flow is not None

def test_get_auth_flow_by_id(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = master_realm.auth_flow(alias='test flow')
    auth_flow_by_id = master_realm.auth_flow(id=auth_flow.id)
    assert auth_flow_by_id is not None

def test_get_none_auth_flow(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = master_realm.auth_flow(alias='not there')
    assert auth_flow is None

def test_get_none_auth_flow_by_id(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = master_realm.auth_flow(id='cdf3b8b6-5cdc-439d-b54a-5d375788af85')
    assert auth_flow is None

def test_get_executions(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='browser').executions()
    assert len(executions) > 0

def test_get_empty_executions(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='test flow').executions()
    assert len(executions) == 0

def test_get_filtered_executions(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='browser').executions(provider='auth-cookie')
    assert len(executions) == 1
    assert executions[0]['displayName'] == 'Cookie'

def test_get_execution(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='browser').executions()
    execution = master_realm.auth_flow(alias='browser').execution(id=executions[0]['id'])
    assert execution is not None


def test_create_execution(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    form_execution = json.loads(json.dumps({'provider': 'auth-username-password-form'}))
    created_execution = master_realm.auth_flow(alias='test flow').create_execution(form_execution)
    assert created_execution is not None


def test_update_execution(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    form_execution = json.loads(json.dumps({'provider': 'auth-username-password-form'}))
    created_execution = master_realm.auth_flow(alias='test flow').create_execution(form_execution)
    assert created_execution.json['requirement'] != 'REQUIRED', "test is brittle, assumes pre-conditions, and fails"
    created_execution.json['requirement'] = 'REQUIRED'
    assert master_realm.auth_flow(alias='test flow').update_execution(created_execution.json).json['requirement'] == 'REQUIRED'

def test_delete_execution(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    executions = master_realm.auth_flow(alias='test flow').executions()
    executions_before_delete = len(executions)
    response = master_realm.auth_flow(alias='test flow').delete_execution(executions[0]['id'])
    assert response.status_code == 204
    assert len(master_realm.auth_flow(alias='test flow').executions()) < executions_before_delete

def test_delete_all_executions(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    master_realm = admin.Admin(session).realm('master')
    auth_flow = {'alias': 'test flow2', 'providerId': 'basic-flow', 'description': 'This flow is used for test purposes', 'topLevel': 'true', 'builtIn': 'false'}
    test_flow2 = admin.Admin(session).realm('master').create_auth_flow(json.loads(json.dumps(auth_flow)))
    test_flow2.create_execution({'provider': 'auth-username-password-form'})
    test_flow2.create_execution({'provider': 'identity-provider-redirector'})
    test_flow2.create_execution({'provider': 'auth-spnego'})
    assert len(master_realm.auth_flow(alias='test flow2').executions()) == 3
    test_flow2.delete_all_executions()
    assert len(master_realm.auth_flow(alias='test flow2').executions()) == 0




from pycloak import merge
import pytest

def test_discrete_value_merge():
    result = merge.merge('wanted', 'unwanted')
    assert result == 'wanted'

def test_discrete_value_merge_None():
    result = merge.merge('wanted', None)
    assert result == 'wanted'

def test_discrete_value_merge_None_preferred():
    result = merge.merge(None, 'unwanted')
    assert result == None

def test_basic_dict_merge_orthogonal():
    result = merge.merge({'a': 1, 'c': 3}, {'b': 2, 'd': 4})
    assert {'a': 1, 'b': 2, 'c': 3, 'd': 4} == result

def test_basic_dict_merge_parallel():
    result = merge.merge({'a': 1, 'c': 3}, {'a': 2, 'c': 4})
    assert {'a': 1, 'c': 3} == result

def test_basic_dict_merge_intersecting():
    result = merge.merge({'a': 1, 'c': 3}, {'a': 2, 'd': 4})
    assert {'a': 1, 'c': 3, 'd': 4} == result

def test_basic_bool_merge_parallel():
    result = merge.merge({'a': True, 'b': False}, {'a': False, 'b': True})
    assert {'a': True, 'b': False} == result

def test_nested_dict_merge_orthogonal():
    preferred = {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32}}
    secondary = {'d': 4, 'e': 5, 'f': {'f1': 61, 'f2': 62}}
    result = merge.merge(preferred, secondary)
    assert {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32}, 'd': 4, 'e': 5, 'f': {'f1': 61, 'f2': 62}} == result

def test_nested_dict_merge_parallel():
    preferred = {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32}}
    secondary = {'a': 4, 'b': 5, 'c': {'c1': 61, 'c2': 62}}
    result = merge.merge(preferred, secondary)
    assert {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32}} == result

def test_nested_dict_merge_intersecting():
    preferred = {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32}}
    secondary = {'a': 4, 'e': 5, 'c': {'c1': 11, 'c3': 33}, 'f': {'f1': 61, 'f2': 62}}
    result = merge.merge(preferred, secondary)
    assert {'a': 1, 'b': 2, 'c': {'c1': 31, 'c2': 32, 'c3': 33}, 'e': 5, 'f': {'f1': 61, 'f2': 62}} == result

def test_nested_dict_merge_orthogonal_child():
    preferred = {'c': {'c1': 31, 'c2': 32}}
    secondary = {'c': {'c3': 33, 'c4': 34}}
    result = merge.merge(preferred, secondary)
    assert {'c': {'c1': 31, 'c2': 32, 'c3': 33, 'c4': 34}} == result

def test_nested_dict_merge_parallel_child():
    preferred = {'c': {'c1': 31, 'c2': 32}}
    secondary = {'c': {'c3': 33, 'c4': 34}}
    result = merge.merge(preferred, secondary)
    assert {'c': {'c1': 31, 'c2': 32, 'c3': 33, 'c4': 34}} == result

def test_nested_dict_merge_intersecting_child():
    preferred = {'c': {'c1': 31, 'c2': 32}}
    secondary = {'c': {'c1': 13, 'c4': 34}}
    result = merge.merge(preferred, secondary)
    assert {'c': {'c1': 31, 'c2': 32, 'c4': 34}} == result

def test_mismatched_type_merge_dict():
    preferred = {'c': {'c1': 31, 'c2': 32}}
    secondary = {'c': [13, 34]}
    result = merge.merge(preferred, secondary)
    assert {'c': {'c1': 31, 'c2': 32}} == result

def test_mismatched_type_merge_list():
    preferred = {'c': [13, 34]}
    secondary = {'c': {'c1': 31, 'c2': 32}}
    result = merge.merge(preferred, secondary)
    assert {'c': [13, 34]} == result

def test_nested_list_merge_orthogonal():
    preferred = {'a': 1, 'b': [21, 22, 23]}
    secondary = {'c': 3, 'd': [41, 42, 43]}
    result = merge.merge(preferred, secondary)
    assert {'a': 1, 'b': [21, 22, 23], 'c': 3, 'd': [41, 42, 43]} == result

def test_nested_list_merge_parallel():
    preferred = {'b': [21, 22, 23]}
    secondary = {'b': [41, 42, 43]}
    result = merge.merge(preferred, secondary)
    assert {'b': [21, 22, 23]} == result




import pytest

@pytest.fixture(scope='session')
def admin_username():
    return 'admin'

@pytest.fixture(scope='session')
def admin_password():
    return 'password'

@pytest.fixture(scope='session')
def keycloak_server(request, admin_username, admin_password):
    print('\n\nStarting docker container for Keycloak server...')
    import docker
    import requests
    import time

    docker_instance_name = 'pycloak-test-server'
    client = docker.from_env()
    client.containers.run('jboss/keycloak:3.4.1.Final', environment={'KEYCLOAK_USER': admin_username, 'KEYCLOAK_PASSWORD': admin_password}, ports={'8080/tcp': '8080'}, name=docker_instance_name, remove=True, detach=True)

    booted = False;
    while not booted:
        print('Waiting for container to boot...')
        time.sleep(1)

        try:
            booted = requests.get('http://127.0.0.1:8080/auth/admin/master/console/').status_code == 200
            print('Keycloak console is up, fixture is ready.')
        except requests.exceptions.ConnectionError:
            print('Keycloak admin console not available')

    def fin():
        print('\n\nStopping docker container for Keycloak server....')
        # spin down docker container
        pycloak_test_severs = client.containers.list(filters={'name': docker_instance_name})
        if pycloak_test_severs:
            pycloak_test_severs[0].stop()
            print('Keycloak server shutdown complete')
        else:
            print('Test server has already been shut down')

    request.addfinalizer(fin)




import json
from pycloak import merge

# TODO - could use some constant definitions for built-in protocols.
# However, don't enumerate since protocols are extensible in Keycloak

class Client:

    def __init__(self, auth_session, json_rep=None, dict_rep=None):
        if not json_rep and dict_rep:
            json_rep = json.loads(json.dumps(dict_rep))

        self.auth_session = auth_session
        self.id = json_rep.get('id')
        self.json = json_rep

    def merge(self, other, prefer_self=False):
        if prefer_self:
            return Client(self.auth_session, dict_rep=merge.merge(self.json, other.json))
        else:
            return Client(self.auth_session, dict_rep=merge.merge(other.json, self.json))




import json
import logging
import requests
from pycloak import realm


class Admin:

    def __init__(self, auth_session):
        self.auth_session = auth_session

    def get_realms(self):
        realms_url = "{0}/auth/admin/realms".format(self.auth_session.host)
        realms_response = requests.get(realms_url, headers=self.auth_session.bearer_header)

        if (realms_response.status_code != 200):
            raise AdminException('Could not retrieve realms')

        return json.loads(realms_response.text)

    realms = property(get_realms)

    def realm(self, realm_name):
        realms_url = "{0}/auth/admin/realms/{1}".format(self.auth_session.host, realm_name)
        realm_response = requests.get(realms_url, headers=self.auth_session.bearer_header)

        if (realm_response.status_code == 404):
            return None
        elif (realm_response.status_code != 200):
            raise AdminException("Could not retrieve realm {}".format(realm_name))

        return realm.Realm(self.auth_session, json_rep=json.loads(realm_response.text))

    def add_realm(self, realm_id, enabled=True):
        realms_url = "{0}/auth/admin/realms".format(self.auth_session.host)
        add_realm_data = {'id': realm_id, 'realm': realm_id, 'enabled': enabled}
        realm_response = requests.post(realms_url, json=add_realm_data, headers=self.auth_session.bearer_header)

        if (realm_response.status_code != 201):
            raise AdminException("Could not add realm {}".format(realm_id))

        # realm has been added, now retrieve
        realm_response = requests.get(
            realm_response.headers['Location'], headers=self.auth_session.bearer_header)
        if (realm_response.status_code != 200):
            raise AdminException("Could not retrieve newly created realm {}".format(realm_name))

        return realm.Realm(self.auth_session, json_rep=json.loads(realm_response.text))

    def update_realm(self, realm):
        realms_url = "{0}/auth/admin/realms/{1}".format(self.auth_session.host, realm.id)
        realm_response = requests.put(realms_url, json=realm.json, headers=self.auth_session.bearer_header)

        if (realm_response.status_code != 204):
            raise AdminException("Could not add realm {}".format(realm_id))

        return self.realm(realm.id)

    def delete_realm(self, realm_id):
        realms_url = "{0}/auth/admin/realms/{1}".format(self.auth_session.host, realm_id)
        realm_response = requests.delete(realms_url, headers=self.auth_session.bearer_header)

        if (realm_response.status_code != 204):
            raise AdminException("Could not delete realm {}".format(realm_id))

class AdminException(Exception):
    pass




from pycloak import auth, admin
import pytest
import json


def test_get_fed_providers(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    fed_providers = admin.Admin(session).realm('master').federation_providers()
    assert len(fed_providers) == 0


def test_add_fed_provider(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    provider_to_add = {'name': 'Test Kerb Provider', 'providerId': 'kerberos',
                       'providerType': 'org.keycloak.storage.UserStorageProvider', 'parentId': 'master'}
    provider_to_add['config'] = {'priority': [0], 'kerberosRealm': ['TESTKERB.COM'],
                                 'serverPrincipal': ['HTTP/test@TESTKERB.COM'], 'keytab': ['/etc/krb5.keytab'], 'debug': ['false'], 'allowPasswordAuthentication': ['false']}
    provider_json = json.loads(json.dumps(provider_to_add))
    new_provider = admin.Admin(session).realm('master').add_federation_provider(provider_json)
    assert new_provider is not None

    fed_providers = admin.Admin(session).realm('master').federation_providers()
    assert len(fed_providers) == 1


def test_get_fed_provider_by_id(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)

    fed_providers = admin.Admin(session).realm('master').federation_providers()
    assert len(fed_providers) == 1, "Unanticipated number of federation providers.  Brittle test is unusable."

    fed_provider_by_id = admin.Admin(session).realm('master').federation_provider(id=fed_providers[0]['id'])
    assert fed_provider_by_id is not None


def test_get_fed_provider_by_name(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    fed_provider_by_name = admin.Admin(session).realm('master').federation_provider(name='Test Kerb Provider')
    assert fed_provider_by_name is not None


def test_update_fed_provider(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kerb_provider = admin.Admin(session).realm('master').federation_provider(name='Test Kerb Provider')
    kerb_provider['config']['debug'] = ['true']

    updated_kerb_provider = admin.Admin(session).realm('master').update_federation_provider(kerb_provider)
    assert updated_kerb_provider['config']['debug'] == ['true']




import requests
import json
from pycloak import execution

class Flow:

    def __init__(self, realm, json_rep=None, dict_rep=None):
        """
        Instantiates a representation of a Keycloak authentication flow.

        :param realm: pycloak realm object to which this flow belongs.
        :param json_rep: json representation of the authentication flow.  Given priority over dict_rep if both are provided.
        :param dict_rep: dictionary representation of the authentication flow.  Discarded if json_rep is provided.
        """
        if not json_rep and dict_rep:
            json_rep = json.loads(json.dumps(dict_rep))

        self.realm = realm
        self.id = json_rep.get('id')
        self.json = json_rep

    # TODO this just occurred to me... the '$host/auth' path can be changed.  should be parameterized!
    # Might be useful to make static url-maker methods in all the respective classes for this?
    def executions(self, provider=None):
        """
        Retrieves list of all executions for this authentication flow

        :return: json list of executions associated with this authentication flow
        """
        # TODO see if the ID can be used instead, no idea why KC console decided to use alias for REST path for executions
        # http://localhost:8080/auth/admin/realms/master/authentication/flows/Test Flow/executions
        executions_url = "{0}/auth/admin/realms/{1}/authentication/flows/{2}/executions".format(self.realm.auth_session.host, self.realm.id, self.json['alias'])
        response = requests.get(executions_url, headers=self.realm.auth_session.bearer_header)

        if response.status_code != 200:
            raise AuthFlowException("could not retrieve executions for auth flow")

        executions = json.loads(response.text)
        if provider is None:
            return executions
        else:
            return list(filter(lambda execution: execution.get('providerId') == provider, executions))

    def execution(self, id=None):
        """
        Retrieves a single execution by ID.

        :param id: GUID of the execution.  I.E. 2cd14612-4901-4d94-81ca-a86729d63fab
        :return: pycloak object of the matching execution.  None if not found or no id is provided.
        """
        executions = self.executions()
        if id is not None:
            execu = next(filter(lambda execu: execu['id'] == id, self.executions()), None)
            return execution.Execution(self, json_rep=execu)

        return None


    def create_execution(self, new_execution):
        """
        Creates an execution in this flow.  Note that the execution will be added according to default Keycloak behavior and
        further customization is likely to be necessary.

        :param execution: json object representing the execution to be added to the flow.  I.E. {'provider': 'auth-username-password-form'}
        :return: arbitrary execution that shares the same provider type.  Most commonly, there will only be one execution of each type,
        so this will be reliable.  However, since Keycloak does not return a Location header on 204, there is no deterministic
        way to validate the correct execution is retrieved.  If something crazy happens None might be returned.  Just warning you.

        Will PR for this soon.
        """
        # TODO validate required "provider" field present

        url = "{0}/auth/admin/realms/{1}/authentication/flows/{2}/executions/execution".format(self.realm.auth_session.host, self.realm.id, self.json['alias'])
        response = requests.post(url, json=new_execution, headers=self.realm.auth_session.bearer_header)

        # TODO keycloak PR for this, return at least a UID or something...
        if response.status_code != 204:
            raise AuthFlowException("Error attempting to create new execution: {0}/{1}".format(response.status_code, response.text))

        best_guess = next(iter(self.executions(provider=new_execution['provider'])), None)
        if best_guess is None:
            return None
        else:
            return execution.Execution(self, json_rep=best_guess)

    def update_execution(self, execution):
        """
        Performs an update operation for an execution, matched by ID

        :return: updated execution object, as retrieved from Keycloak
        """

        url = "{0}/auth/admin/realms/{1}/authentication/flows/{2}/executions".format(self.realm.auth_session.host, self.realm.id, self.json['alias'])
        response = requests.put(url, json=execution, headers=self.realm.auth_session.bearer_header)

        if response.status_code != 204:
            raise AuthFlowException("Error attempting to update execution: {0}/{1}".format(response.status_code, response.text))

        return self.execution(id=execution['id'])

    def delete_execution(self, id):
        """
        Deletes execution specified by ID.  If not found, no error is raised.
        """
        url = "{0}/auth/admin/realms/{1}/authentication/executions/{2}".format(self.realm.auth_session.host, self.realm.id, id)
        response = requests.delete(url, headers=self.realm.auth_session.bearer_header)

        if response.status_code not in [404, 204]:
            raise AuthFlowException("Error attempting to delete execution")

        #TODO feels wrong - return something else here
        return response

    def delete_all_executions(self):
        """
        Convenience method for remove all executions associated with this flow

        :return: fresh flow object with executions removed.
        """
        for execution in self.executions():
            self.delete_execution(execution['id'])

        return self.realm.auth_flow(id=self.id)


    def create_child_flow(self, child_flow):
        """
        Creates child flow, with the current flow as parent

        :param child_flow: json representation of the flow to create.  I.E. {'alias': 'test-sub-flow', 'type': 'basic-flow', 'description': 'Demonstrates how to configure a sub-flow', 'provider': 'registration-page-form'}
        """
        url = "{0}/auth/admin/realms/{1}/authentication/flows/{2}/executions/flow".format(self.realm.auth_session.host, self.realm.id, self.json['alias'])
        response = requests.post(url, json=child_flow, headers=self.realm.auth_session.bearer_header)

        if response.status_code != 204:
            raise AuthFlowException("Error attempting to create new execution")

        return next(iter(self.executions(provider=execution['provider'])))

    def update_child_flow(self, flow):
        """
        Just invokes update_execution.  Since both child flows and executions share the same structure, they use the same REST invocations
        """
        self.update_execution(flow)

class AuthFlowException(Exception):
    pass




from pycloak import auth, admin, realm, client
import pytest

def test_auth_success(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    access_token = session.get_access_token()
    assert access_token != None

def test_auth_failure(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, 'NOT' + admin_password)
    with pytest.raises(auth.AuthException, message='Invalid credentials did not cause exception'):
        access_token = session.get_access_token()

def test_get_offline_token(keycloak_server, admin_username, admin_password):
    # First, have to make sure that admin-cli has access to all roles, otherwise offline_token requests will fail
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    admin_cli = kc_admin.realm('master').client_id('admin-cli')
    admin_cli.json['fullScopeAllowed'] = 'true'
    kc_admin.realm('master').update_client(admin_cli.json)

    token_response = auth.direct_access_grant_token(admin_username, admin_password, include_offline=True)
    assert token_response.get('refresh_token') != None

def test_use_offline_token(keycloak_server, admin_username, admin_password):
    direct_access_token_response = auth.direct_access_grant_token(admin_username, admin_password, include_offline=True)
    offline_token_response = auth.offline_token(direct_access_token_response['refresh_token'])
    assert offline_token_response.get('access_token') != None

def test_create_session_offline_token(keycloak_server, admin_username, admin_password):
    direct_access_token_response = auth.direct_access_grant_token(admin_username, admin_password, include_offline=True)
    session = auth.AuthSession(offline_token=direct_access_token_response['refresh_token'])
    assert session.get_access_token() is not None




from pycloak import auth, admin, realm
import pytest

def test_list_realms(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    realms = kc_admin.realms
    assert len(realms) != 0, "No realms returned by /realms endpoint"

def test_get_realm(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    master = kc_admin.realm('master')
    assert master.id == 'master', 'master realm not found'

def test_get_realm_not_found(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    nonexistent_realm = kc_admin.realm('none')
    assert nonexistent_realm is None

def test_add_realm(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    new_realm = kc_admin.add_realm('add-realm-test')
    assert new_realm.id == 'add-realm-test', 'could not successfully add new realm'

def test_update_realm(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    new_realm = kc_admin.add_realm('update-realm-test')
    assert new_realm.id == 'update-realm-test', 'could not successfully add new realm for update test'
    assert new_realm.json['sslRequired'] != 'none', 'sslRequired flag already set to "none", cannot perform update test'
    new_realm.json['sslRequired'] = 'none'
    updated_realm = kc_admin.update_realm(new_realm)
    assert updated_realm.json['sslRequired'] == 'none'

def test_delete_realm(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    new_realm = kc_admin.add_realm('delete-realm-test')
    assert new_realm.id == 'delete-realm-test', 'could not successfully add new realm for delete test'
    kc_admin.delete_realm('delete-realm-test')
    assert kc_admin.realm('delete_realm_test') is None

def test_merge_realm(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    new_realm = kc_admin.add_realm('merge-realm-test')
    assert new_realm.id == 'merge-realm-test', 'could not successfully add new realm for update test'
    merge_realm = realm.Realm(session, dict_rep={'id': 'merge-realm-test', 'accessCodeLifespan': 33})
    merge_result = new_realm.merge(merge_realm)
    assert merge_result.json['accessCodeLifespan'] == 33
    assert merge_result.json['enabled'] == True

def test_merge_realm_preferring_self(keycloak_server, admin_username, admin_password):
    session = auth.AuthSession(admin_username, admin_password)
    kc_admin = admin.Admin(session)
    new_realm = kc_admin.add_realm('merge-realm-test2')
    assert new_realm.id == 'merge-realm-test2', 'could not successfully add new realm for update test'
    merge_realm = realm.Realm(session, dict_rep={'id': 'merge-realm-test2', 'accessCodeLifespan': 33})
    merge_result = new_realm.merge(merge_realm, prefer_self=True)
    assert merge_result.json['accessCodeLifespan'] != 33




import json
import jwt
import logging
import requests
import time
import urllib.parse
import urllib.request

logger = logging.getLogger('pycloak.auth')


def direct_access_grant_token(username, password, host="http://localhost:8080", realm="master", client_id="admin-cli", include_offline=False):
    """
    Makes a request against the Keycloak openid-connect/token endpoint using a direct access grant to get a token.

    Returned JSON objects will look like this:
    {'access_token': 'ey...2w', 'expires_in': 60, 'refresh_expires_in': 1800, 'refresh_token': 'ey..Og', 'not-before-policy': 0, 'session_state': '5ca9ccf7-b425-4fe8-81a1-8bbd770ee312', 'token_type': 'bearer'}

    @param username: user for which the token is to be granted
    @param password: user's password
    @param host: full hostname with protocol of the keycloak auth server
    @param realm: realm against which the direct access grant will be made
    @param client_id: client argument for the direct access grant request
    @param include_offline: denotes whether or not to include the parameter for an offline token request.
    @return: json object representing token Response
    @raise: AuthException if a non-200 response is returned by the token endpoint
    """
    token_request_data = {'grant_type': 'password', 'client_id': client_id, 'username': username, 'password': password}
    if include_offline:
        token_request_data['scope'] = 'offline_access'
    token_request_url = "{0}/auth/realms/{1}/protocol/openid-connect/token".format(host, realm)
    token_response = requests.post(token_request_url, data=token_request_data)

    if token_response.status_code != 200:
        raise AuthException('Non-200 Response returned from direct access grant attempt: {}'.format(token_response.status_code))

    return json.loads(token_response.text)

def offline_token(offline_token , host="http://localhost:8080", realm="master", client_id="admin-cli", include_offline=False):
    token_request_data = {'grant_type': 'refresh_token', 'client_id': client_id, 'refresh_token': offline_token}

    if include_offline:
        token_request_data['scope'] = 'offline_access'
    token_request_url = "{0}/auth/realms/{1}/protocol/openid-connect/token".format(host, realm)
    token_response = requests.post(token_request_url, data=token_request_data)

    if token_response.status_code == 400:
        raise AuthException('400/Bad Request returned.  Validate that the client allows this kind of grant and the calling user has the offline_access role (and it is effective for the client)')
    if token_response.status_code != 200:
        raise AuthException('Non-200 Response returned from offline token refresh attempt: {}'.format(token_response.status_code))

    return json.loads(token_response.text)

def is_expired(token):
    """
    Takes a base-64 encoded representation of a JSON web token and determines if it is expired.  Note that this is the *only* check it performs.

    @param token: base-64 encoded representation of JWT
    @return: true if expired, false if non-expired
    """
    decoded_jwt = jwt.decode(token, verify=False)
    try:
        exp = int(decoded_jwt['exp'])
    except ValueError:
        logging.error('Could not determine token expiry using "exp" claim.  Pleave validate that valid tokens are being returned from the endpoint.')
        raise AuthException("Could not determine token expiry")

    return exp < int(time.time())

class AuthSession:

    __access_token = None

    def __init__(self, username=None, password=None, offline_token=None, host="http://localhost:8080", realm="master", client_id="admin-cli"):
        if offline_token is None and (username is None or password is None):
            raise AuthException("Unable to authenticate, no form of credentials provided.  Please supply either a username/password pair or offline token.")

        self.username = username
        self.password = password
        self.token = offline_token
        self.host = host
        self.realm = realm
        self.client_id = client_id

    #TODO should be using refresh tokens instead of direct access grants for re-auth every time
    def get_access_token(self):
        if self.__access_token is None or is_expired(self.__access_token):
            if self.token is not None:
                logging.debug('Using offline token grant to acquire access token.')
                token_response = offline_token(self.token, self.host, self.realm, self.client_id)
            else:
                logging.debug('Using direct access grant to acquire access token.')
                token_response = direct_access_grant_token(self.username, self.password, self.host, self.realm, self.client_id)

            self.__access_token = token_response['access_token']

        return self.__access_token

    def bearer_header(self):
        return {'Authorization': "Bearer {}".format(self.get_access_token())}

    access_token = property(get_access_token)
    bearer_header = property(bearer_header)

class AuthException(Exception):

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



from distutils.core import setup

setup(
    name='pycloak',
    packages=['pycloak','pycloak.test'],
    version='0.3.1',
    description='API for Keycloak Admin Interface',
    author='Josh Cain',
    author_email='jcain@redhat.com',
    license='GNU Lesser General Public License, V3',
    long_description=open('README.txt').read(),
)




import json
import logging
import requests
from pycloak import client, merge, flow


class Realm:

    def __init__(self, auth_session, json_rep=None, dict_rep=None):
        if not json_rep and dict_rep:
            json_rep = json.loads(json.dumps(dict_rep))

        self.auth_session = auth_session
        self.id = json_rep.get('id')
        self.json = json_rep

    # TODO add an optional filter criteria to clients
    def clients(self):
        clients_url = "{0}/auth/admin/realms/{1}/clients".format(self.auth_session.host, self.id)
        clients_response = requests.get(clients_url, headers=self.auth_session.bearer_header, params={'viewableOnly': True})

        # TODO only return none on not found, all other erros should exception
        if (clients_response.status_code != 200):
            return None

        return json.loads(clients_response.text)

    def client(self, id):
        client_url = "{0}/auth/admin/realms/{1}/clients/{2}".format(self.auth_session.host, self.id, id)
        client_response = requests.get(client_url, headers=self.auth_session.bearer_header)

        if (client_response.status_code != 200):
            raise RealmException(
                "Could not retrieve clients for id: {}.  Did you specify using the clientId field instead of GUID?".format(id))

        return client.Client(self.auth_session, json_rep=json.loads(client_response.text))

    def client_id(self, client_id):
        client_json = next(filter(lambda client: client.get('clientId') == client_id, self.clients()), None)
        if client_json:
            return client.Client(self.auth_session, json_rep=client_json)
        else:
            return None

    def create_client(self, client_id, protocol, rootUrl=None):
        create_client_url = "{0}/auth/admin/realms/{1}/clients".format(
            self.auth_session.host, self.id)
        create_client_response = requests.post(create_client_url, json={'enabled': True, 'attributes': {
        }, 'redirectUris': [], 'clientId': client_id, 'protocol': protocol, 'rootUrl': rootUrl}, headers=self.auth_session.bearer_header)

        # TODO fix inconsistent errors... most things just throw an exception
        # instead of providing useful feedback
        if create_client_response.status_code != 201:
            logging.error("Failed to create client: [{}]{}".format(
                create_client_response.status_code, create_client_response.text))
            raise RealmException("Could not create client with clientId: {}".format(client_id))

        new_client = requests.get(create_client_response.headers['Location'], headers=self.auth_session.bearer_header)
        if new_client.status_code != 200:
            raise RealmException("Error attempting to retrieve newly created client: {}".format(client_id))

        return client.Client(self.auth_session, json_rep=json.loads(new_client.text))

    def update_client(self, updated_client):
        client_url = "{0}/auth/admin/realms/{1}/clients/{2}".format(self.auth_session.host, self.id, updated_client.get('id'))
        update_client_response = requests.put(client_url, json=updated_client, headers=self.auth_session.bearer_header)

        if update_client_response.status_code != 204:
            raise RealmException("Error attempting to update client: {}".format(updated_client['id']))

        updated_client = requests.get(client_url, headers=self.auth_session.bearer_header)
        if updated_client.status_code != 200:
            raise RealmException("Error attempting to retrieve newly created client: {}".format(client_id))

        return client.Client(self.auth_session, json_rep=json.loads(updated_client.text))

    def delete_client(self, id):
        client_url = "{0}/auth/admin/realms/{1}/clients/{2}".format(self.auth_session.host, self.id, id)
        delete_client_response = requests.delete(client_url, headers=self.auth_session.bearer_header)

        if delete_client_response.status_code != 204:
            raise RealmException("Error attempting to delete client: {}".format(id))

    def merge(self, other, prefer_self=False):
        if prefer_self:
            return Realm(self.auth_session, dict_rep=merge.merge(self.json, other.json))
        else:
            return Realm(self.auth_session, dict_rep=merge.merge(other.json, self.json))

    def federation_providers(self):
        fed_provider_url = "{0}/auth/admin/realms/{1}/components?parent={1}&type=org.keycloak.storage.UserStorageProvider".format(
            self.auth_session.host, self.id)
        get_fed_providers_response = requests.get(fed_provider_url, headers=self.auth_session.bearer_header)

        if get_fed_providers_response.status_code != 200:
            raise RealmException("Error attempting to retrieve federation providers for realm: {}".format(self.id))

        return json.loads(get_fed_providers_response.text)

    # TODO inconsistent verbiage + method names.  Going too fast, need to slow down and standardize
    def add_federation_provider(self, federation_provider):
        fed_provider_url = "{0}/auth/admin/realms/{1}/components".format(self.auth_session.host, self.id)
        add_fed_provider_response = requests.post(fed_provider_url, json=federation_provider,
                                                  headers=self.auth_session.bearer_header)

        if add_fed_provider_response.status_code != 201:
            raise RealmException("Error attempting to create federation provider: {0}.  Recieved HTTP code: {1}, with message: {2}".format(
                federation_provider['name'], add_fed_provider_response.status_code, add_fed_provider_response.text))

        retrieve_new_fed_provider_response = requests.get(add_fed_provider_response.headers[
                                                          'Location'], headers=self.auth_session.bearer_header)
        if retrieve_new_fed_provider_response.status_code != 200:
            raise RealmException("Error attempting to retrieve newly created federation provider: {}".format(
                federation_provider['name']))

        return json.loads(retrieve_new_fed_provider_response.text)

    def federation_provider(self, id=None, name=None):
        if id is not None:
            fed_provider_url = "{0}/auth/admin/realms/{1}/components/{2}".format(self.auth_session.host, self.id, id)
            get_fed_provider_response = requests.get(fed_provider_url, headers=self.auth_session.bearer_header)

            # TODO only return none on 404/Not Found
            if get_fed_provider_response.status_code != 200:
                return None

            return json.loads(get_fed_provider_response.text)

        elif name is not None:
            return next(filter(lambda provider: provider.get('name') == name, self.federation_providers()), None)

        else:
            raise RealmException(
                "Cannot lookup federation provider for realm {} without an ID or name.  Please specify one.".format(self.id))

    def update_federation_provider(self, updated_provider):
        fed_provider_url = "{0}/auth/admin/realms/{1}/components/{2}".format(
            self.auth_session.host, self.id, updated_provider['id'])
        update_fed_provider_response = requests.put(
            fed_provider_url, json=updated_provider, headers=self.auth_session.bearer_header)

        if update_fed_provider_response.status_code != 204:
            raise RealmException("Unable to update federation provider with id: {0}.  Returned HTTP {1}, with message: {2}".format(
                updated_provider['id'], update_fed_provider_response.status_code, update_fed_provider_response.text))

        return self.federation_provider(id=updated_provider['id'])

    # TODO fed provider delete

    def auth_flows(self):
        """
        Lists authentication flows for the given realm

        :returns: json list of all authentication flows for the realm
        """
        auth_flows_url = "{0}/auth/admin/realms/{1}/authentication/flows".format(self.auth_session.host, self.id)
        get_auth_flows_response = requests.get(auth_flows_url, headers=self.auth_session.bearer_header)

        if get_auth_flows_response.status_code != 200:
            raise RealmException("Error attempting to retrieve authentication flows for realm: {}".format(self.id))

        # TODO make this OO? - is inconsistent
        return json.loads(get_auth_flows_response.text)

    # TODO generalize bad status code error handling.  I'm certain that there's a construct that will make this faster (perhaps 'with')?
    def create_auth_flow(self, auth_flow):
        """
        Creates a new authentication flow

        :param auth_flow: json object representing the new auth flow.  Ex: {"alias":"Test Flow","providerId":"basic-flow","description":"This flow is to test creation functions","topLevel":true,"builtIn":false}
        :returns: pycloak object representation of the newly created flow
        """
        auth_flows_url = "{0}/auth/admin/realms/{1}/authentication/flows".format(self.auth_session.host, self.id)
        create_auth_flow_response = requests.post(auth_flows_url, json=auth_flow, headers=self.auth_session.bearer_header)

        if create_auth_flow_response.status_code != 201:
            raise RealmException("Error attempting to create new auth flow: {0}/{1}".format(create_auth_flow_response.status_code, create_auth_flow_response.text))

        if create_auth_flow_response.headers.get('Location'):
            get_new_flow_response = requests.get(create_auth_flow_response.headers['Location'], headers=self.auth_session.bearer_header)
            if get_new_flow_response.status_code != 200:
                raise RealmException("Error attempting to retrieve newly created auth flow: {0}/{1}".format(get_new_flow_response.status_code, get_new_flow_response.text))

            return flow.Flow(self, json_rep=json.loads(get_new_flow_response.text))
            
        # Older versions don't return a 'Location' header, so do this the wrong way.
        else:
            return self.auth_flow(alias=auth_flow['alias'])

    def auth_flow(self, id=None, alias=None):
        """
        Retrieves a pycloak auth flow object by id or alias.  If both are given, id will be used.

        :param id: GUID of the auth flow (I.E. "cdf3b8b6-5cdc-439d-b54a-5d375788af85")
        :param alias: friendly name of the auth flow (I.E. "browser")
        :return: pycloak object representing an auth flow
        """
        if id is not None:
            auth_flow_url = "{0}/auth/admin/realms/{1}/authentication/flows/{2}".format(self.auth_session.host, self.id, id)
            auth_flow_response = requests.get(auth_flow_url, headers=self.auth_session.bearer_header)

            if auth_flow_response.status_code == 404:
                return None
            elif auth_flow_response.status_code == 200:
                return flow.Flow(self, json_rep=json.loads(auth_flow_response.text))
            else:
                raise RealmException("Error attempting to retrieve auth flow by ID")

        elif alias is not None:
            flow_json = next(filter(lambda flow: flow.get('alias') == alias, self.auth_flows()), None)
            if flow_json:
                return flow.Flow(self, json_rep=flow_json)

        # fall-through case
        return None

    def auth_config(self, id):
        url = "{0}/auth/admin/realms/{1}/authentication/config/{2}".format(self.auth_session.host, self.id, id)
        response = requests.get(url, headers=self.auth_session.bearer_header)

        if response.status_code == 404:
            return None
        elif response.status_code == 200:
            return json.loads(response.text)
        else:
            raise RealmException("Error occurred attempting to retrieve config")

    # TODO general notes to self:
    #  - on all lookup operations (GET), return None for not found status code, object if found, exception otherwise.  Enforce this library-wide

class RealmException(Exception):
    pass
